#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# -------------------------------------------------------------------------------------
#
# Copyright (c) 2020 Skinnyhouse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:
#
#    - Redistributions of source code must retain the above copyright notice, this 
#    list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above copyright notice, 
#    this list of conditions and the following disclaimer in the documentation and/or 
#    other materials provided with the distribution.
#    - Neither the name of the author nor the names of its contributors may be used 
#    to endorse or promote products derived from this software without specific prior 
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
# SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
# DAMAGE.
#
# -------------------------------------------------------------------------------------
#
# This file is a basic Python plug-in for GIMP.
# It was written after receiving a request from Pat625 on GimpLearn
# It can be executed by selecting the menu option: 'Filters>Custom Font Tools/Custom Font Paragrapher
#
# The filter uses the function "computeThetaWithSlope" written by Ofnuts
# and parts of code taken from Tranductin's filter, custom-font_on_path, which he attributes to a post on http://stackoverflow.com/questions/31735499/calculate-angle-clockwise-between-two-points
# Without their work I would not have been able to create this filter
#
# Version 001 Takes the user text from the filter dialog and writes it along a series of horizontal paths
#             Options provided for setting custom font size, margin sizes, line spacing and background fill
#             Copes with accented characters
#             Replaces characters that do not exist with space

# Version 002 Removed 500 character limit from the input string: leading spaces stripped from text
#             Added option to place text along existing user path
#             Re-worked check for missing character layers
#             Increased range for line spacing in filter dialog

# Version 003 Fixed missing newline character
#             Fixed missing character when one long string of characters is entered

# Version 004 Added options for indenting paragraphs

# Version 005 Added centre layers option

# Version 006 Improved handling of kerning paths
# Version 007 Improved handling of kerning paths



from gimpfu import *
#from array import array
#import math
#import random

#from gtk import gdk
#w = gdk.screen_width()
#h = gdk.screen_height()

# Some global colours we may need to define

black = gimpcolor.RGB(0, 0, 0)
white = gimpcolor.RGB(255, 255, 255)
#darkgrey = gimpcolor.RGB(50, 50, 50)
#midgrey = gimpcolor.RGB(125, 125, 125)
#lightgrey = gimpcolor.RGB(170, 170, 170)
#verylightgrey = gimpcolor.RGB(222, 222, 222)
#red = gimpcolor.RGB(220,10,10)	

# Some global metallic colours we may need for creating gradients

#lightsilver = gimpcolor.RGB(204, 202, 202)
#darksilver = gimpcolor.RGB(46, 43, 43)
#lightgold = gimpcolor.RGB(250, 233, 49)
#darkgold = gimpcolor.RGB(240, 183, 42)
#lightcopper = gimpcolor.RGB(212, 125, 1)
#darkcopper = gimpcolor.RGB(177, 106, 58)
#lightbrass = gimpcolor.RGB(253, 242, 91)
#darkbrass = gimpcolor.RGB(220 , 118, 20)
#lightbronze = gimpcolor.RGB(104, 75, 33)
#darkbronze = gimpcolor.RGB(83, 69, 42)


def do_stuff(image, layer, xcffilename, text, font_size, line_spacing, top_margin, left_margin, right_margin, bottom_margin, indent_type, indent_size, background_type, drop_offset_x, drop_offset_y, drop_blur, drop_opacity, drop_colour, use_paths, centred):
    #### Main Filter Proc
	
    # Store the GIMP's settings so they can be restored
    gimp.context_push()
    # Set start undo point
    pdb.gimp_image_undo_group_start(image)	
    text = text.decode('utf-8')
    dummy = 0
    custom_font_test, fontimg = fn_test_valid_custom_font(xcffilename)
    if custom_font_test:
        original_text = text

        add_drop = False
        if drop_colour>0: add_drop = True
		
        missing_layers = False
        font_name_pointer, file_dummy = -1, 0
        fontimg = pdb.gimp_xcf_load(file_dummy,xcffilename,xcffilename)
        text, missing_layers, missing_amount, missing_text = fn_check_xcf_text(text,fontimg, missing_layers)
		
        image.active_layer =  layer
        loop, copy_vector, copy_vector_name = 0, -1, ""
        kern_amount, use_custom, horizontal_padding, font_name = 0, True, 0, "Arial"
		
        if use_paths:            
            num_vectors, vector_ids = pdb.gimp_image_get_vectors(image)
            vector_list = image.vectors
            if num_vectors == 0: use_paths = False
		
        if use_paths:
            # Get the current active path i.e. the newly created temp path
            copy_vector = pdb.gimp_image_get_active_vectors(image)
            if copy_vector == -1:
                pdb.gimp_image_set_active_vectors(image, vector_ids[0])# use the first path for the active path
                copy_vector = pdb.gimp_image_get_active_vectors(image)
                copy_vector_name = pdb.gimp_item_get_name(copy_vector)
            copy_vector_name = pdb.gimp_item_get_name(copy_vector)
	
            # Get path stroke information
            NumberOfStrokes_in_Path,ActivePath_StrokeIDs_List = pdb.gimp_vectors_get_strokes(copy_vector)
            ActivePath_StrokeID = ActivePath_StrokeIDs_List[0]
            copy_vector_length = pdb.gimp_vectors_stroke_get_length(copy_vector,ActivePath_StrokeID,0.01)
			
            #pdb.gimp_message("copy_vector_name = " + copy_vector_name +  " : copy_vector_length = " + str(copy_vector_length) )
			
            # temp code below to produce an output
            path_length = image.width - (right_margin + left_margin)
            if path_length<(image.width/4): path_length = int(image.width/4)
			
            # Added 13/04
            path_length = copy_vector_length
			
            lined_text_list, newline_list = fn_get_text_for_para_lines(image, text, font_name, font_size, kern_amount, use_custom, xcffilename, fontimg, copy_vector_length, horizontal_padding, use_paths, copy_vector, copy_vector_name, indent_type, indent_size)
			
        else:
            # since we are using only horizontal lines here we can work out the length of the path from the image width and margin sizes
            path_length = image.width - (right_margin + left_margin)
            if path_length<(image.width/4): path_length = int(image.width/4)
				
            lined_text_list, newline_list = fn_get_text_for_para_lines(image, text, font_name, font_size, kern_amount, use_custom, xcffilename, fontimg, path_length, horizontal_padding, use_paths, copy_vector, copy_vector_name, indent_type, indent_size)
			
        # monitor text output
        #pdb.gimp_message(lined_text_list)
        #pdb.gimp_message(newline_list)
		
        # variables required by proc_text_along_path
        use_custom_font = True
        kern_amount = 0
        step_size = 0
        do_best_size = False
	
        font_name_pointer = -1
        fg_colour = pdb.gimp_context_get_foreground()
        blend_colour = pdb.gimp_context_get_background()
        use_gradient_colour = False
        spread_colour = False
        font_colour = fg_colour
        gradient_colour = blend_colour
        keep_layers = False
        keep_paths = False
        use_space_to_change_step_size = False
        compensate_multiplier = 0
        use_text_size_modifier = False
        initial_text_size = 0
        medial_text_size = 0
        final_text_size = 0
        
        position_of_medial = 0
        do_repeat = False
        remaining_path = 0
        reversed_writing = False
        Left_to_Right_path = False
	
        jitter = False
        jitter_range = 0
        tranparency = 100
        reversed_vector = -1
        custom_font_mode = 47 # number 22 in original mode selection menu
	
        rotate_text = False
        user_angle = 0
		
        # variables required by proc_text_along_path but nor by paragrapher
        vertical = False
        folder_info = ""
        vertical_adjustment_list = ""
        horizontal_padding = 0
        vertical_padding = 0
        diagonal = False
         
		
        if len(lined_text_list)>1:
            
            if background_type>0:
                bg_layer = pdb.gimp_layer_new(image, image.width, image.height, 1, "background layer", 100, 28)
                pdb.gimp_image_insert_layer(image, bg_layer, None, -1)
                if background_type == 1: 
                    pdb.gimp_drawable_fill(bg_layer, 0)
                elif background_type == 2:
                    pdb.gimp_drawable_fill(bg_layer, 1)
            loop = 1 # start with the second item in the list
            while loop<len(lined_text_list):			
			
		
                # variables required by proc_text_along_path
                #font_size = 100
                use_custom_font = True
                kern_amount = 0
                step_size = 0
                do_best_size = False
	
                font_name_pointer = -1
                fg_colour = pdb.gimp_context_get_foreground()
                blend_colour = pdb.gimp_context_get_background()
                use_gradient_colour = False
                spread_colour = False
                font_colour = fg_colour
                gradient_colour = blend_colour
                keep_layers = False
                keep_paths = False
                use_space_to_change_step_size = False
                compensate_multiplier = 0
                use_text_size_modifier = False
                initial_text_size = 0
                medial_text_size = 0
                final_text_size = 0
        
                position_of_medial = 0
                do_repeat = False
                remaining_path = 0
                reversed_writing = False
                Left_to_Right_path = False
	
                jitter = False
                jitter_range = 0
                tranparency = 100
                reversed_vector = -1
                custom_font_mode = 47 # number 22 in original mode selection menu
	
                rotate_text = False
                user_angle = 0
		
                # variables required by proc_text_along_path but nor by paragrapher
                vertical = False
                folder_info = ""
                vertical_adjustment_list = ""
                horizontal_padding = 0
                vertical_padding = 0
                diagonal = False
			
                if use_paths:
                    # copy and re-position existing path as line guide
                    new_path, current_y_pos = fn_copy_line_guide_path(image, loop-1, font_size, line_spacing, copy_vector, copy_vector_name)
                else:
                    # Draw a straight path for the text to follow
                    new_path, current_y_pos = fn_draw_line_guide_path(image, loop-1, font_size, top_margin, left_margin, right_margin, line_spacing)
		
                # Get the current active path i.e. the newly created temp path
                active_vector = pdb.gimp_image_get_active_vectors(image)
                active_vector_name = pdb.gimp_item_get_name(active_vector)
			
                # Place the custom font text on the path
                final_temp_text, text = lined_text_list[loop], lined_text_list[loop]
				
                indent = False
                if indent_type>0:
                    indent = newline_list[loop-1]
                    #if loop == 1: indent = True
                    #if lined_text_list[loop-1] == "" or loop == 1:
                    #    indent = True
                #if loop>2: pdb.gimp_message("indent = " + str(indent) + " : text = " + str(lined_text_list[loop-1]))
				
                custom_scale = proc_text_along_path(image, layer, final_temp_text, text, font_name_pointer,
                                    font_name, font_colour, blend_colour, spread_colour, gradient_colour, use_gradient_colour, font_size, step_size, kern_amount, 
                                    keep_layers, keep_paths, use_space_to_change_step_size, 
                                    compensate_multiplier, use_text_size_modifier, initial_text_size, medial_text_size, final_text_size,  do_best_size,
                                    position_of_medial, final_temp_text, do_repeat, remaining_path, reversed_writing, Left_to_Right_path, rotate_text, user_angle, 
                                    jitter, jitter_range, tranparency, use_custom_font, xcffilename, fontimg, reversed_vector, custom_font_mode, path_length, vertical, 
                                    folder_info, vertical_adjustment_list, horizontal_padding, vertical_padding, diagonal, indent, indent_type, indent_size)
			
			
                loop += 1
				
			
            # Check to see if the image and background layer need re-sizing to accommodate the lines of text 
            if use_paths:
                active_layer = pdb.gimp_image_get_active_layer(image)
                offset_x, offset_y = pdb.gimp_drawable_offsets(active_layer)
                current_y_pos = offset_y + (active_layer.height)
            if current_y_pos + (font_size/2) + bottom_margin > image.height: 
                pdb.gimp_image_resize(image, image.width, current_y_pos + (font_size/2) + bottom_margin, 0, 0)
                if background_type>0: pdb.gimp_layer_resize_to_image_size(bg_layer)
                if background_type == 1: 
                    pdb.gimp_drawable_fill(bg_layer, 0)
                elif background_type == 2:
                    pdb.gimp_drawable_fill(bg_layer, 1)
					
            # Add a basic drop shadow effect to text layers?
            if add_drop: 
                end_drop = (len(lined_text_list))
                loop = 0
                num_layers, layer_ids = pdb.gimp_image_get_layers(image)
                layer_names_list = image.layers
                while loop < end_drop:
                    drop_layer = layer_names_list[loop]
                    drop_name = pdb.gimp_item_get_name(drop_layer)
                    if "Custom Font" in drop_name:
                        pdb.gimp_image_set_active_layer(image, drop_layer)
                        proc_dropshadow(image, drop_layer, drop_offset_x, drop_offset_y, drop_blur, drop_opacity, drop_colour)
                    loop += 1
					
					
            # Use centred justification for text layers
            if centred: 
                end_loop = (len(lined_text_list))
                loop = 0
                num_layers, layer_ids = pdb.gimp_image_get_layers(image)
                layer_names_list = image.layers
                while loop < end_loop:
                    centred_layer = layer_names_list[loop]
                    centred_name = pdb.gimp_item_get_name(centred_layer)
                    if "Custom Font" in centred_name:
                        #pdb.gimp_image_set_active_layer(image, drop_layer)
                        offset_x, offset_y = pdb.gimp_drawable_offsets(centred_layer)
                        pdb.gimp_layer_set_offsets(centred_layer,int((image.width-centred_layer.width)/2),offset_y)
                    loop += 1
					
            # Remove paths from image	
            # get lists of vector ids in image
            num_vectors, vector_ids = pdb.gimp_image_get_vectors(image)
            vector_list = image.vectors
            loop = 0
            while loop<num_vectors:
                vector_id = vector_list[loop]
                path_name = pdb.gimp_item_get_name(vector_list[loop])
                path_name = path_name.decode('utf-8')
                if "skinny line guide path " in path_name: pdb.gimp_image_remove_vectors(image, vector_list[loop])
                loop += 1

    else:
        pdb.gimp_message("Please select a custom font before you continue.")
        
	
    # Set end undo point
    pdb.gimp_image_undo_group_end(image)

    # Restore the GIMP's settings
    gimp.context_pop()
	
######################################################################################################################################

def fn_copy_line_guide_path(image, loop, font_size, line_spacing, copy_vector, copy_vector_name):
    new_vector = pdb.gimp_vectors_copy(copy_vector)
    pdb.gimp_image_insert_vectors(image, new_vector, None, -1)
    new_vector_name = "skinny line guide path " + str((loop + 1))
    pdb.gimp_item_set_name(new_vector, new_vector_name)
    pdb.gimp_item_transform_translate(new_vector, 0 , loop * (line_spacing + font_size))
    line_y_pos = (loop * (line_spacing + font_size)) # temp value
    return new_vector_name, line_y_pos

def fn_draw_line_guide_path(image, loop, font_size, top_margin, left_margin, right_margin, line_spacing):
    if loop == 0:
        line_y_pos = ((font_size / 2) + top_margin)
    else:
        line_y_pos = ((font_size / 2) + top_margin) + ((font_size + line_spacing) * loop)
        #line_y_pos = ((loop) * font_size) + (font_size / 2) + top_margin + line_spacing
    line_x0_pos = left_margin
    line_x1_pos = image.width - right_margin
    
    new_vector_name = "skinny line guide path " + str((loop + 1))
    new_vector_id = pdb.gimp_vectors_new(image, new_vector_name)
    pdb.gimp_image_insert_vectors(image, new_vector_id, None, -1)
    points1 = [line_x0_pos,line_y_pos] * 3
    points2 = [line_x1_pos,line_y_pos] * 3
    points = points1 + points2
    new_vector_id = pdb.gimp_vectors_stroke_new_from_points(new_vector_id ,0, len(points), points, False)
    return new_vector_name, line_y_pos

#######################################################################################################

def fn_font_x_pos_adjustment(kerning, f_size):
    adjust_factor = (((100 * kerning) / 1024) * f_size) / 100
    return adjust_factor
	
######################################################################################################################################

def fn_check_text_is_valid(text, fontimg):
    # checks that characters are present in the custom font, replacing any missing characters with a space
    loop, new_text = 0,""
    while loop<len(text):
        char = text[loop]
        if char == "'": pdb.gimp_message("Apostrophe detected")
        
        if char<>"\n":
            layer_name = char + ".png"
            #pdb.gimp_message(layer_name)
            char_layer = pdb.gimp_image_get_layer_by_name(fontimg,layer_name)
            pdb.gimp_message(layer_name + " : " + str(char_layer) )
            if char_layer < 0:
                if char.lower() == char:
                    char = char.upper()
                else:
                    char = char.lower()
                layer_name = char + ".png"
                char_layer = pdb.gimp_image_get_layer_by_name(fontimg,layer_name)
                if char_layer < 0:
                    char = " "
        new_text += char
           
        loop += 1
		
    return new_text

######################################################################################################################################


def fn_test_valid_custom_font(xcffilename):
    test = False
    fontimg = -1
    if xcffilename > "":
        try:
            fontimg = pdb.gimp_xcf_load(0,xcffilename,xcffilename)
        except RuntimeError:
            fontimg = -1
		
        else:
            fontimg = pdb.gimp_xcf_load(0,xcffilename,xcffilename)
            fontimg_layer = pdb.gimp_image_get_layer_by_name(fontimg, "space.png")
            if fontimg_layer >= 0: test = True
    return test, fontimg

#######################################################################################################

def fn_label_characters(label_text,label_len):
    label_char_info = label_text
    label_len_info = len(label_text)
    label_etc = ""
    if label_len_info > label_len: 
        label_char_info = label_char_info[0:label_len]
        if label_char_info.count(" ")>0:
            loop = 0
            space_pos = label_len - 1
            while loop<label_len:
                if label_char_info[loop] == " ": space_pos = loop
                loop += 1
            label_char_info = label_char_info[0:space_pos]
            label_etc = " etc"
    if label_len_info == 1: 
        label_len_detail = " char" 
    else:
        label_len_detail = " chars" 
    details = label_char_info + label_etc + " " + str(label_len_info ) + label_len_detail
    return details
	
######################################################################################################################################

def fn_get_text_for_para_lines(image, text, f_name, f_size, kern_amount, use_custom, xcffilename, fontimg, path_length, horizontal_padding, use_paths, active_vector, active_vector_name, indent_type, indent_size):

		
    #text = text.decode('utf-8') # Commented out temp 30/03
    font_name_pointer = -1
    #f_size = f_size_seed
    char_width, NextChr_width = 0,0
    kerning_flag = -1
    kerning_vector = -1
    # Calculate kerning to be used between characters, if required by user, based on size of font taken from user settings
    kern_gap = kern_amount
    tempbox_width = 0
    loop = 0
    padding = 0
    f_xpos = 0
    init_f_xpos = f_xpos
    scaled_xpos = 0
    curve = 0
	
    space_name = " .png"
    char_name = " "
    layer_name = char_name + ".png"
	
	
    if use_custom:
        char_layer = pdb.gimp_image_get_layer_by_name(fontimg,layer_name)
        font_name_pointer = -1
        space_has_vectors = False
        if char_layer < 0: #layer not found
            space_name = "space.png"
            char_name = "space"
            layer_name = char_name + ".png"
            char_layer = pdb.gimp_image_get_layer_by_name(fontimg,layer_name)
        if char_layer < 0: #layer not found
            space_name = "nospace"
            #pdb.gimp_message("No space character..." )
            space_width = f_size * 1.0 # Make up a space width for this font size
            space_height = f_size
            font_y_offset = (f_size * 1.0)/2
        else:
            custom_scale = fn_get_custom_font_scale(f_size, char_layer.height)
            space_width = char_layer.width * custom_scale
            space_height = char_layer.height * custom_scale
            font_y_offset = space_height/2
    else:
	
        # Find the width to use for the space character
        (fontbox_width, fontbox_height, fontbox_ascent, fontbox_descent) = pdb.gimp_text_get_extents_fontname(" ", f_size, PIXELS, str(f_name))
        space_width, space_height = fontbox_width, fontbox_height
        #### should we check for paths/glyphs in the space character
		
        # Check to see if the space character has paths so that we can set a flag if necessary
        temp_space_string = " -"
        temp_space_layer = pdb.gimp_text_layer_new(image,temp_space_string,f_name,f_size,0)
        pdb.gimp_image_insert_layer(image,temp_space_layer,None,-1)
        temp = pdb.gimp_vectors_new_from_text_layer(image, temp_space_layer)
        space_vectors_path = pdb.gimp_image_insert_vectors(image,temp,None,-1)
        space_vectors = pdb.gimp_image_get_vectors_by_name(image, temp_space_string)
        # Get space character path stroke information
        NumberOfStrokes_in_Space_Path,Space_Path_StrokeIDs_List = pdb.gimp_vectors_get_strokes(space_vectors)
        pdb.gimp_image_remove_vectors(image,space_vectors)
        temp = pdb.gimp_image_remove_layer(image, temp_space_layer)

        # Check to see if the font uses a (visible) glyph in the space character to set a flag
        space_has_vectors = False
        if (NumberOfStrokes_in_Space_Path > 1):
            if (font_name_pointer > -1): space_has_vectors = True
			
	
    kern_gap = ((float(space_width)/100))*horizontal_padding
    if kern_gap < 0: kern_gap = 0
    
    # define paragrapher specific variables
    lined_text_list = ["dummy"]
    newline_list = [True]
    lined_text_list_loop = 1
    space_pos = -1
    current_line_text = ""
    temp_counter = 0
	
    indent_width = 0
    if indent_type>0: 
        indent_width = space_width * indent_size
        f_xpos = indent_width
    init_f_xpos = f_xpos
    scaled_xpos = f_xpos
	
    while loop < len(text):
        #### put check for space character here????
		
        CurrentChar = text[loop]
		
        if CurrentChar == "\n": 
            #lined_text_list_loop+=1
            if current_line_text>"":
                current_line_text = current_line_text.lstrip()
                lined_text_list.append(current_line_text)
                f_xpos = indent_width
                newline_list.append(True)
            else:
                lined_text_list.append("")
                newline_list.append(True)
                
            #pdb.gimp_message("Newline encountered")
            current_line_text = ""
            space_pos = -1
            
            #lined_text_list.append(current_line_text)
			
        else:
		
            if use_custom:
                ch_layer = pdb.gimp_image_get_layer_by_name(fontimg,space_name)
                ch_scale = fn_get_custom_font_scale(f_size, ch_layer.height)
                sp_width = ch_layer.width * ch_scale
			
            # Find the width to use for the space character
            if CurrentChar == " ":
                if current_line_text == "": # for Paragrapher
                    space_pos = 0
                else:
                    space_pos = len(current_line_text)-1
				
                if use_custom:
                    char_layer = pdb.gimp_image_get_layer_by_name(fontimg,space_name)
                    custom_scale = fn_get_custom_font_scale(f_size, char_layer.height)
                    space_width = char_layer.width * custom_scale
                    custom_width = space_width
                    scaled_width = space_width
                else:
                    (fontbox_width, fontbox_height, fontbox_ascent, fontbox_descent) = pdb.gimp_text_get_extents_fontname(" ", f_size, PIXELS, str(f_name))
                    space_width = fontbox_width
                    fontbox_width = space_width
                    scaled_width = space_width
		
            if loop < (len(text)-1):
                NextChar = text[loop + 1]
            else:
                NextChar = ""
			
			
			
            # reset character x position adjustment factors
            xpos_adjustment = 0
		
            if (CurrentChar > " " or (CurrentChar == " " and space_has_vectors == True)):
                custom_layer_name = CurrentChar.decode('utf-8') + ".png"
                # Create layers for each character in user defined string	

                if use_custom:
                    char = CurrentChar.decode('utf-8')
                    # get id of character from custom font file in memory
                    #pdb.gimp_message(str(custom_layer_name.decode('utf-8')))
                    char_layer = pdb.gimp_image_get_layer_by_name(fontimg, custom_layer_name.decode('utf-8'))
                    char_vector = pdb.gimp_image_get_vectors_by_name(fontimg, custom_layer_name.decode('utf-8'))
                    kerning_name = CurrentChar # Added 30/03
                    if char_layer < 0: #if letter not found try swapping upper/lower case to look for
                        #char = CurrentChar
                        if char.islower():
                            char = char.upper()
                            custom_layer_name = char + ".png"
                        elif char.isupper():
                            char = char.lower()
                            custom_layer_name = char + ".png"
                        char_layer = pdb.gimp_image_get_layer_by_name(fontimg, custom_layer_name.decode('utf-8'))
                        char_vector = pdb.gimp_image_get_vectors_by_name(fontimg, custom_layer_name.decode('utf-8'))
                        kerning_name = char # Added 30/03
                    kerning_name2 = kerning_name + ".png" # Added 30/03
				
                    if NextChar<>"":
                        nchar = NextChar
                        nchar_custom_layer_name = nchar.decode('utf-8')+".png" # Added 30/03
                        nchar_layer = pdb.gimp_image_get_layer_by_name(fontimg, nchar_custom_layer_name)
                        if nchar_layer < 0: #if letter not found try swapping upper/lower case to look for
                            if nchar.islower():
                                nchar = nchar.upper()
                            elif nchar.isupper():
                                nchar = nchar.lower()
                            nchar_layer = pdb.gimp_image_get_layer_by_name(fontimg, nchar.decode('utf-8')+".png")
                            #temp_kerning_name = kerning_name + nchar # Added 30/03
                        if nchar_layer>-1:
                            kerning_pair_name = char + nchar + ".png"
                            kerning_pair_vector = pdb.gimp_image_get_vectors_by_name(fontimg, kerning_pair_name.decode('utf-8'))
                            if kerning_pair_vector>-1: char_vector = kerning_pair_vector
							
                        #kerning_name = temp_kerning_name # Added 30/03
                        #kerning_name_2 = kerning_name + ".png" # Added 30/03
                    ####
                    # sort out scaling to get the font size we need
                    custom_scale = fn_get_custom_font_scale(f_size, char_layer.height)
                    custom_width = (char_layer.width  + kern_gap) * custom_scale
                    custom_height = char_layer.height * custom_scale
				
				
                    scaled_width = custom_width

                    ####
                    # create, insert and set visibility the copy of the character layer into our image
                else:
                    #### New 0.10 lines
                    font_scale = fn_get_custom_font_scale(f_size, fontbox_height)
                    char_layer = pdb.gimp_text_layer_new(image,CurrentChar,f_name,f_size,0)
				
                    fontbox_width = char_layer.width #+ (kern_gap * font_scale)
                    scaled_width = char_layer.width
				
				
	            #### Look for kerning path information
                #kerning_vector = -1
                #char_vector = -1
                if use_custom:
                    #char_vector = pdb.gimp_image_get_vectors_by_name(fontimg,custom_layer_name)
                    #kerning_name, kerning_name_2 = CurrentChar + NextChar, CurrentChar + NextChar + ".png" # Commented out 30/03
                    #kerning_vector = pdb.gimp_image_get_vectors_by_name(fontimg,kerning_name) # check for character kerning pair paths without .png
                    #if kerning_vector == None: 
                    #    kerning_vector = pdb.gimp_image_get_vectors_by_name(fontimg,kerning_name_2) # check for character kerning pair paths with .png
                    
                    #if char_vector>-1 or kerning_vector>-1: kerning_flag = 0   
                    if char_vector>-1: kerning_flag = 0   
                    #pdb.gimp_message("kerning_vector = " + str(kerning_vector) )					
					
                if char_vector >=0:
                    # Get path info for current character path to work out character width
                    #char_vector_name = pdb.gimp_item_get_name(char_vector)
                    NumberOfStrokes_in_Path,Char_Vector_StrokeIDs_List = pdb.gimp_vectors_get_strokes(char_vector)
                    Char_Vector_Stroke = Char_Vector_StrokeIDs_List[0]
                    TypeOfStrokeInChar_Vector,NumberOfPointsInChar_Vector,ListOfControlPoints,IsStrokeClosed = pdb.gimp_vectors_stroke_get_points(char_vector,Char_Vector_Stroke)
                    firstX = ListOfControlPoints[2]
                    secondX = ListOfControlPoints[len(ListOfControlPoints)-4]
                    minX = min(firstX,secondX)
                    maxX = max(firstX,secondX)
                    vector_char_width = maxX - minX #overwrite char_width to be used as character width.
                    add_width = vector_char_width
                    if use_custom: custom_width = add_width * custom_scale ##################### altered for loads of text
                    if use_custom: f_xpos += (minX * custom_scale) # Slab v 0.06 caters for negative kerning with custom fonts ##################### altered for loads of text
				
                '''
                if kerning_vector >= 0: 
                    NumberOfStrokes_in_Path,Char_Vector_StrokeIDs_List = pdb.gimp_vectors_get_strokes(kerning_vector)
                    Char_Vector_Stroke = Char_Vector_StrokeIDs_List[0]
                    TypeOfStrokeInChar_Vector,NumberOfPointsInChar_Vector,ListOfControlPoints,IsStrokeClosed = pdb.gimp_vectors_stroke_get_points(kerning_vector,Char_Vector_Stroke)
                    firstX = ListOfControlPoints[2]
                    secondX = ListOfControlPoints[len(ListOfControlPoints)-4]
                    minX = min(firstX,secondX)
                    maxX = max(firstX,secondX)
                    kerning_width = maxX - minX #overwrite char_width to be used as character width.
                    add_width = kerning_width
                    if use_custom: custom_width = add_width * custom_scale ##################### altered for loads of text
                    if use_custom: f_xpos += (minX * custom_scale) # Slab v 0.06 caters for negative kerning with custom fonts ##################### altered for loads of text
                    '''
            #### Update f_xpos to take width of text along path from
            if use_custom:
                if loop == len(text)-1 and kerning_flag == 0: 
					
                    custom_width = custom_width + (custom_width * 0.5)
                    scaled_width = scaled_width + (scaled_width * 0.5)					
					
				
                scaled_xpos = scaled_xpos + scaled_width + curve - xpos_adjustment
                #f_xpos = f_xpos + custom_width + kern_gap + curve - xpos_adjustment	
                f_xpos = f_xpos + custom_width + curve - xpos_adjustment
			
                if current_line_text == "": f_xpos += (f_xpos/2) # For Paragrapher
			
                if not loop == len(text)-1: f_xpos = f_xpos + abs(kern_gap)
                #if f_xpos<0: f_xpos = 0
			
                if f_xpos < init_f_xpos: f_xpos = init_f_xpos
			
            else:
                if loop == len(text)-1: fontbox_width = fontbox_width + fontbox_width * 0.5
                f_xpos = f_xpos + fontbox_width + kern_gap + curve - xpos_adjustment
                if loop == len(text)-1: 
                    scaled_width = scaled_width + (scaled_width * 0.5)
                scaled_xpos = scaled_xpos + scaled_width + curve - xpos_adjustment
			
            current_line_text += CurrentChar # for Paragrapher
		
            #if f_xpos>path_length and loop<len(text):# for Paragrapher
            if f_xpos >(path_length - (custom_width * 0.5)) and loop<len(text):# for Paragrapher 14/04
                if space_pos>-1:
                    loop_difference = len(current_line_text)-(space_pos+1)
                    current_line_text = current_line_text[0:space_pos+1]
                    current_line_text = current_line_text.lstrip()
		
                    #pdb.gimp_message("Initial line of text = :" + current_line_text)
                else:
                    loop_difference = 0
                    current_line_text = current_line_text[0:len(current_line_text)]
                loop -= loop_difference
                current_line_text = current_line_text[0:len(current_line_text)]
                lined_text_list.append(current_line_text)
                newline_list.append(False)
                lined_text_list_loop+=1
                space_pos = -1
                current_line_text = ""
                f_xpos = 0
                scaled_xpos = 0
            elif (loop == len(text)-1) and f_xpos<=path_length:
                current_line_text = current_line_text[0:len(current_line_text)]
                current_line_text = current_line_text.lstrip()
                #pdb.gimp_message("Line " + str(loop) + " of text = :" + current_line_text)
                lined_text_list.append(current_line_text)
                newline_list.append(False)
            
			
        loop += 1
		
    original_f_xpos = f_xpos
	
    return lined_text_list, newline_list
	
#######################################################################################################

def fn_check_xcf_text(text_to_check,fontimg, missing_layers):
    # converts any missing xcf-font characters to a space character
    new_text, missing_text, loop, missing_amount, file_dummy = "","",0,0,0
    while loop<len(text_to_check):
        char = text_to_check[loop]
        if char == "\n":
            new_text += char
        else:
            layer_name = char + ".png"
            if char == " ":
                char_layer = pdb.gimp_image_get_layer_by_name(fontimg,layer_name)
                if char_layer < 0:
                    layer_name = "space" + ".png"
        
            char_layer = pdb.gimp_image_get_layer_by_name(fontimg,layer_name)
            if char_layer < 0: #if letter not found try swapping upper/lower case to look for
                if char.lower() == char:
                    char = char.upper()
                else:
                    char = char.lower()
                layer_name = char + ".png"
                
                char_layer = pdb.gimp_image_get_layer_by_name(fontimg,layer_name)
            if char_layer < 0: #if letter still not found
                if loop>0: new_text += " "
                missing_layers = True
                missing_amount += 1
                missing_text += text_to_check[loop] + " "
            else:
                new_text += text_to_check[loop]
        loop += 1
    return new_text, missing_layers, missing_amount, missing_text
	
################################################################  ######################################################################

def fn_get_custom_font_name(filepath):
    search = filepath.rfind("\\")
    part_name = filepath[search + 1:(len(filepath))]
    search = part_name.rfind(".")
    part_name = part_name[0:search]
    return part_name
#######################################################################################################

def fn_get_custom_font_scale(f_size, custom_font_height):
    new_custom_height = f_size * 1.0
    scale = new_custom_height / custom_font_height
    return scale

#######################################################################################################

def computeThetaWithSlope(fromPoint,toPoint,slope):
    dX=toPoint[0]-fromPoint[0]
    dY=toPoint[1]-fromPoint[1]

    if math.fabs(dX) < .001:
        theta=math.copysign(math.pi/2,dY)
    else:
        theta=math.atan(slope)
        if dX<0:
            theta=theta+math.pi
    return theta	
	
#######################################################################################################

def fn_compensate_for_curve_in_path(p_angle, c_angle, f_size):
    curve_compensation,angle_diff = 0,0
	
    if (c_angle<0 and p_angle>180):
        c_angle += 360
    if (c_angle>180 and p_angle<0):
        c_angle -= 360
	
    if (c_angle>p_angle): # tests for text along path travelling in a clockwise direction
        
        min_angle = min(abs(p_angle), abs(c_angle))
        max_angle = max(abs(p_angle), abs(c_angle))
        angle_diff = (min_angle - max_angle)
        curve_compensation = fn_font_x_pos_adjustment(angle_diff, f_size)                

    if (c_angle<p_angle): # tests for  text along path travelling in a counter-clockwise direction
						
        min_angle = min(abs(p_angle), abs(c_angle))
        max_angle = max(abs(p_angle), abs(c_angle))
        angle_diff = max_angle - min_angle
        curve_compensation = fn_font_x_pos_adjustment(angle_diff,f_size)
		
    return curve_compensation,angle_diff

#####################################################################################################
    
def proc_text_along_path(image, layer, joining_text_list, text, font_name_pointer, 
                        f_name, f_colour,  bl_colour, spread_colour, joinedup_gradient, use_gradient_colour, f_size, step_size, kern_amount, 
                        keep_layers, keep_paths, use_space_to_change_step_size, compensate_multiplier, 
                        use_text_size_modifier, initial_text_size, medial_text_size, final_text_size, do_best_size,
                        position_of_medial, final_temp_text, repeat_text, remaining_path, reversed_writing, Left_to_Right_path, rotate_text, 
                        user_angle, jitter, jitter_range, tranparency, use_custom, xcffilename, fontimg, reversed_vector, custom_font_mode, path_length, 
                        vertical, folder_info,vertical_adjustment_list, horizontal_padding, vertical_padding, diagonal, indent, indent_type, indent_size):

				
    # Creates text layers
	
	#### New function to calculate font size that will fit the length of the path
    offset_x = 0
    kerning_vector = -1
    custom_scale = 1
    path_padding = 0
    if do_best_size: f_size, path_padding = fn_size_to_fit_path(image, text, f_name, f_size, kern_amount, use_custom, xcffilename, fontimg)
	
    initial_step_size = step_size
    label_details = fn_label_characters(text, 12)
    label_fontsize_info = " " + str(int(f_size)) + "pt"
    label_fontname = ""
    label_rotate = ""
    label_jitter = ""
    label_reverse = ""
    if use_custom:
        label_fontname = fn_get_custom_font_name(xcffilename)
        label_fontname = "Custom Font " + label_fontname
    else:
        label_fontname = f_name
    
    original_user_angle = user_angle
    governor = 0 # initialise variable to use to prevent creation of unnecessary layers
    attack, decay = 0, 0 # initialise variable to mark point of change in text size within text string when size modifier flag is set
    spaces_loop = 0
    spaces_count = text.count(" ")
	
    if spaces_count > 0:
        if use_space_to_change_step_size:
            if text.endswith(" "): 
                padding = remaining_path / (spaces_count + 0)
            else:
                if remaining_path > 40:
                    padding = remaining_path / (spaces_count + 1)
                else:
                    padding = 0
        else:
            if text.endswith(" "):
                padding = remaining_path / (spaces_count + 0)
            else:
                padding = remaining_path / (spaces_count + 1)
            
    else:
        if font_name_pointer > -1:
            padding = 0
        else:
            if text.endswith(" "):
                padding = remaining_path / len(joining_text_list)
            else:
                padding = remaining_path / (len(joining_text_list) + 1)
						
    if not use_gradient_colour:
        if spaces_count<2: spaces_count = 2
    target = 0
    original_font_colour = f_colour
    shading_loop = 0
	
    if len(text) % 2 == 0: # variable to help position attack (first) phase of change in text size
        even = True
    else:
        even = False
		
    if " " not in text: use_space_to_change_step_size = False # Force option off, if there are no space characters in the text
    #### deleted in v008 down to next quad hash
    if use_text_size_modifier:
        if use_space_to_change_step_size:  
            target = text.count(" ")
            if target<2:
                use_space_to_change_step_size = False
                target = len(text)
                use_text_size_modifier = False #New line
            else:
                if target % 2 == 0:
                    even = True
                else:
                    even = False
        else:
            target = len(text)
            if target<3: use_text_size_modifier = False
    ####
    if use_gradient_colour: 
        # get number of colours etc needed for using colours from current ui gradient
        current_gradient = joinedup_gradient
        if use_space_to_change_step_size:
            list_of_gradient_positions = fn_get_gradient_positions(spaces_count)
            number_of_samples, colours_of_samples = pdb.gimp_gradient_get_custom_samples(current_gradient, len(list_of_gradient_positions), list_of_gradient_positions,False)
        else:
            list_of_gradient_positions = fn_get_gradient_positions(len(final_temp_text))
            number_of_samples, colours_of_samples = pdb.gimp_gradient_get_custom_samples(current_gradient, len(list_of_gradient_positions), list_of_gradient_positions,False)
    #### deleted in v008 down to next quad hash
    if use_text_size_modifier:
        f_size = initial_text_size
        if target == 3: # ie minimum number of characters/spaces to work with
            attack = 1
        else:
            f_size = initial_text_size
            attack = int((float(target)/100) * int(position_of_medial)) # set attack to the point in the text string/space count where the font size will change
            if attack>(target - 1): attack = (target - 1)
	
    if use_text_size_modifier: # get font step size for attack phase
        size_difference,new_step_size = 0,0
        if medial_text_size > initial_text_size: size_difference = medial_text_size - initial_text_size
        if initial_text_size > medial_text_size: size_difference = initial_text_size - medial_text_size
        if target>3:
            if not use_space_to_change_step_size: 
                if even: attack -= 1
            if attack < 1: attack = 1
            new_step_size = size_difference/(attack)
        else:
            new_step_size = size_difference/(attack)
			
        #if new_step_size < 0.5: new_step_size = 0.5
        if medial_text_size < initial_text_size: new_step_size = 0 - new_step_size
        step_size = new_step_size
    ####
    # Get the current active path
    active_vector = pdb.gimp_image_get_active_vectors(image)
    active_vector_name = pdb.gimp_item_get_name(active_vector)
	
    if (active_vector == -1):
        active_vector = pdb.gimp_image_set_active_vectors(image,image.vectors[0])
	
    active_vectors = pdb.gimp_image_get_active_vectors(image)
    # Get path stroke information
    NumberOfStrokes_in_Path,ActivePath_StrokeIDs_List = pdb.gimp_vectors_get_strokes(active_vector)
    ActivePath_StrokeID = ActivePath_StrokeIDs_List[0]
    path_length = pdb.gimp_vectors_stroke_get_length(active_vectors,ActivePath_StrokeID,0.01)
	
	# initialise some variables
    ####f_xpos = path_padding + 0
    f_ypos = 40
    curve = 0
    curve_compensation = 0
    temp_angle_diff = 0
	
    if use_custom:
        
        file_dummy = 0
        space_name = " .png"
        char_name = " "
        layer_name = char_name + ".png"
        char_layer = pdb.gimp_image_get_layer_by_name(fontimg,layer_name)
        if char_layer < 0: #layer not found
            space_name = "space.png"
            char_name = "space"
            layer_name = char_name + ".png"
            char_layer = pdb.gimp_image_get_layer_by_name(fontimg,layer_name)
        if char_layer < 0: #layer not found
            space_name = "nospace"
            #pdb.gimp_message("No space character..." )
            space_width = f_size * 1.164 # Make up a space width for this font size
            space_height = f_size
            custom_width = space_width # Added 30/03
        else:
            custom_scale = fn_get_custom_font_scale(f_size, char_layer.height)
            space_width = char_layer.width * custom_scale
            space_height = char_layer.height * custom_scale
            font_y_offset = (char_layer.height * 1.164)/2
            custom_height = space_height # Added 30/03
            custom_width = space_width # Added 30/03

        kern_gap = ((float(space_width)/100))*horizontal_padding
        
    else:
	
        # Find the width to use for the space character
        (fontbox_width, fontbox_height, fontbox_ascent, fontbox_descent) = pdb.gimp_text_get_extents_fontname(" ", f_size, PIXELS, str(f_name))
        space_width = fontbox_width
		
        kern_gap = ((float(space_width)/100))*horizontal_padding
	
        # Check to see if the space character has paths so that we can set a flag if necessary
        temp_space_string = " -"
        temp_space_layer = pdb.gimp_text_layer_new(image,temp_space_string,f_name,f_size,0)
        pdb.gimp_image_insert_layer(image,temp_space_layer,None,-1)
        temp = pdb.gimp_vectors_new_from_text_layer(image, temp_space_layer)
        space_vectors_path = pdb.gimp_image_insert_vectors(image,temp,None,-1)
        space_vectors = pdb.gimp_image_get_vectors_by_name(image, temp_space_string)
        # Get space character path stroke information
        NumberOfStrokes_in_Space_Path,Space_Path_StrokeIDs_List = pdb.gimp_vectors_get_strokes(space_vectors)
        pdb.gimp_image_remove_vectors(image,space_vectors)
        temp = pdb.gimp_image_remove_layer(image, temp_space_layer)

        # Check to see if the font uses a (visible) glyph in the space character to set a flag
        space_has_vectors = False
        if (NumberOfStrokes_in_Space_Path > 1):
            if (font_name_pointer > -1): space_has_vectors = True
			
		
    indent_width = 0
    if indent and indent_type>0:
        indent_width = indent_size * space_width
		
		
    f_xpos = indent_width + path_padding + 0
    temp_f_xpos = f_xpos
    init_f_xpos = f_xpos

    text_layer_y_origin = image.height - (f_size + f_ypos)
			
    initial_f_size = f_size		
    if initial_f_size < 2: initial_f_size = 2
    if initial_f_size > 3000: initial_f_size = 3000
    
    if not use_custom:
        (fontbox_width, fontbox_height, fontbox_ascent, fontbox_descent) = pdb.gimp_text_get_extents_fontname("dummy", initial_f_size, PIXELS, str(f_name))
        font_y_offset = fontbox_ascent

    char_width, NextChr_width = 0,0
    # Calculate kerning to be used between characters, if required by user, based on size of font taken from user settings
    kern_factor = kern_amount/f_size
    #kern_gap = kern_factor * f_size
    tempbox_width = 0
    loop = 0
	
    if use_custom: 
        font_name_pointer = -1
        space_has_vectors = False
	
    # Initialise a loop specifically to use for creating layer and path identifiers and names
    layer_and_path_loop = 0
			        
    while loop < len(joining_text_list) and (f_xpos + tempbox_width < path_length):
        CurrentChar, RealCurrentChar = joining_text_list[loop], text[loop]
        tempbox_width = 0

        if loop > 0:		        
            if " " not in joining_text_list:
                f_xpos += padding
            else:
                # Find the width to use for the space character
                if CurrentChar == " ":
                    if use_custom:
                        char_layer = pdb.gimp_image_get_layer_by_name(fontimg,space_name)
                        custom_scale = fn_get_custom_font_scale(f_size, char_layer.height)
                        space_width = char_layer.width * custom_scale
                        custom_width = space_width
                    else:
                        (fontbox_width, fontbox_height, fontbox_ascent, fontbox_descent) = pdb.gimp_text_get_extents_fontname(" ", f_size, PIXELS, str(f_name))
                        space_width = fontbox_width
                    f_xpos += padding
        if loop < (len(joining_text_list)-1):
            NextChar, RealNextChar = joining_text_list[loop + 1], text[loop + 1]
        else:
            NextChar, RealNextChar = "", ""
			
        # reset character x position adjustment factors
        f_xpos_adjustment, j_xpos_adjustment, p_xpos_adjustment, t_xpos_adjustment = 0,0,0,0
        xpos_adjustment = 0

        if (font_name_pointer > -1):
            # Adjust kerning
            if reversed_writing:
                if Left_to_Right_path:
                    xpos_adjustment += fn_adjust_xpos(NextChar, CurrentChar, RealCurrentChar, RealNextChar, font_name_pointer, f_size, joining_text_list[loop])
				 
                else:
                    
                    xpos_adjustment += fn_adjust_xpos(NextChar, CurrentChar, RealCurrentChar, RealNextChar, font_name_pointer, f_size, joining_text_list[loop])
            else:
                xpos_adjustment += fn_adjust_xpos(NextChar, CurrentChar, RealCurrentChar, RealNextChar, font_name_pointer, f_size, joining_text_list[loop])
				
        if (CurrentChar > " " or (CurrentChar == " " and space_has_vectors == True)):
            custom_layer_name = CurrentChar + ".png"
            # Create layers for each character in user defined string		        
            if (loop == 0):
				
                if use_custom:
                    # get id of character from custom font file in memory
                    char_layer = pdb.gimp_image_get_layer_by_name(fontimg, custom_layer_name)
                    if NextChar<>"": n_char_layer = pdb.gimp_image_get_layer_by_name(fontimg, NextChar+".png")
                    char = CurrentChar
                    if char_layer < 0: #if letter not found try swapping upper/lower case to look for
                        char = CurrentChar
                        if char.lower() == char:
                            char = char.upper()
                        else:
                            char = char.lower()
                        custom_layer_name = char + ".png"
                        char_layer = pdb.gimp_image_get_layer_by_name(fontimg, custom_layer_name)
                    n_char = NextChar
                    if NextChar<>"":
                        n_char_layer = pdb.gimp_image_get_layer_by_name(fontimg, NextChar+".png")
                        if n_char_layer < 0: #if letter not found try swapping upper/lower case to look for
                            n_char = NextChar
                            if n_char.lower() == n_char:
                                n_char = n_char.upper()
                            else:
                                n_char = n_char.lower()
						
                    # work out scale of character 
                    custom_scale = fn_get_custom_font_scale(f_size, char_layer.height)
                    # create, insert and set visibility the copy of the character layer into our image
                    copy = pdb.gimp_layer_new_from_drawable(char_layer,image)
                    pdb.gimp_image_insert_layer(image,copy,None,0)
                    pdb.gimp_item_set_visible(copy,True)
                    # sort out scaling to get the font size we need
                    custom_width = char_layer.width * custom_scale
                    custom_height = copy.height * custom_scale
                    old_interpolation = pdb.gimp_context_get_interpolation() # Note current interpolation method
                    pdb.gimp_context_set_interpolation(3) # Ensure that interpolation is set to Sinc (Lancos 3) method
                    pdb.gimp_layer_scale(copy,custom_width,custom_height,False)
                    pdb.gimp_context_set_interpolation(old_interpolation) # Restore old interpolation method
                    pdb.gimp_layer_set_offsets(copy,0,0)
                    # Create a list to hold the id of the first character
                    layer_id_list = [copy]
                    # Create a list to hold the layer name
                    layer_name_list = [CurrentChar] # Was: layer_name_list = [custom_layer_name]
					
                    # Set the new custom layer colour from the current font colour and the colour to shade to
                    if spread_colour and not use_gradient_colour: 
                        if not use_space_to_change_step_size:
                            custom_colour = original_font_colour
                        else:
                            custom_colour = original_font_colour
                        proc_colour_character(image, layer_id_list[layer_and_path_loop], custom_colour, custom_font_mode)
                        # Re-set the id of the layer
                        layer_id_list = [pdb.gimp_image_get_active_layer(image)]
                    # Set the new text layer colour from the current gradient sampled colour
                    if use_gradient_colour:
                        custom_colour = fn_shade_gradient_colour(colours_of_samples,loop)
                        proc_colour_character(image, layer_id_list[layer_and_path_loop], custom_colour, custom_font_mode)
                        # Re-set the id of the layer
                        layer_id_list = [pdb.gimp_image_get_active_layer(image)]
                else:
                    # Create a list to hold the text layer identifiers
                    layer_id_list = ["text_layer" + str(layer_and_path_loop)]
                    # Create a list to hold the text layer names
                    layer_name_list = [str(joining_text_list[layer_and_path_loop])]
				
                    # Create the layer for the first character
                    layer_id_list[layer_and_path_loop] = pdb.gimp_text_layer_new(image,CurrentChar,f_name,f_size,0)
				
                    # and insert this first layer into the image
                    pdb.gimp_image_insert_layer(image,layer_id_list[layer_and_path_loop],None,-1)
				
                    # Set the new text layer colour
                    if use_gradient_colour: f_colour = fn_shade_gradient_colour(colours_of_samples,loop)
                    pdb.gimp_text_layer_set_color(layer_id_list[layer_and_path_loop], f_colour)
					
                # Set layer opacity as per user setting
                pdb.gimp_layer_set_opacity(layer_id_list[layer_and_path_loop], tranparency)
				
                # Rename the new layer 
                pdb.gimp_item_set_name(layer_id_list[layer_and_path_loop],str(text[loop]))

                # Set the x,y position of the first layer within the image
                pdb.gimp_layer_translate(layer_id_list[layer_and_path_loop],f_xpos,text_layer_y_origin)

                # note the width of the current character
                char_width = layer_id_list[layer_and_path_loop].width
				
                offset_x,offset_y = pdb.gimp_drawable_offsets(layer_id_list[layer_and_path_loop])

                # To insert paths of letters in the text list
                if use_custom:
                    dummy = 0
                    # get the id of the path of the layer from the custom font
                    char_vector = pdb.gimp_image_get_vectors_by_name(fontimg,custom_layer_name)
                    # Create a list to hold the path layer ids
                    path_id_list = [char_vector]
                    # Create another list to hold the names of the paths
                    paths_names_list = [str(CurrentChar)]
                else:
                    # Create a list to hold the path layer ids
                    path_id_list = [pdb.gimp_vectors_new_from_text_layer(image,layer_id_list[layer_and_path_loop])]
                    # Create another list to hold the names of the paths
                    paths_names_list = [str(joining_text_list[layer_and_path_loop])]

                    # Add the path to the image
                    temp = pdb.gimp_image_insert_vectors(image,path_id_list[layer_and_path_loop],None,-1)
                    # Get the path id associated with the current character
                    char_vector = pdb.gimp_image_get_vectors_by_name(image, paths_names_list[layer_and_path_loop])
	
                kerning_vector = -1
                if use_custom:
                    #kerning_name, kerning_name_2 = CurrentChar + NextChar, CurrentChar + NextChar + ".png"
                    kerning_name, kerning_name_2 = char + n_char, char + n_char + ".png"
                    kerning_vector = pdb.gimp_image_get_vectors_by_name(fontimg,kerning_name) # check for character kerning pair paths without .png
                    if kerning_vector == None: kerning_vector = pdb.gimp_image_get_vectors_by_name(fontimg,kerning_name_2) # check for character kerning pair paths with .png
                    #if kerning_vector == None: pdb.gimp_message("No kerning path for kerning pair " + str(kerning_name_2))
					
                if char_vector >=0 and kerning_vector <0:
                    # Get path info for current character path to work out character width
                    # Next two temp lines not necessary since this is the first character
                    #char_vector_name = pdb.gimp_item_get_name(char_vector)
                    #pdb.gimp_message(" Vector name = " + str(char_vector_name) )
                    NumberOfStrokes_in_Path,Char_Vector_StrokeIDs_List = pdb.gimp_vectors_get_strokes(char_vector)
                    Char_Vector_Stroke = Char_Vector_StrokeIDs_List[0]
                    TypeOfStrokeInChar_Vector,NumberOfPointsInChar_Vector,ListOfControlPoints,IsStrokeClosed = pdb.gimp_vectors_stroke_get_points(char_vector,Char_Vector_Stroke)
                    firstX = ListOfControlPoints[2]
                    secondX = ListOfControlPoints[len(ListOfControlPoints)-4]
                    minX = min(firstX,secondX)
                    maxX = max(firstX,secondX)
                    vector_char_width = maxX - minX #overwrite char_width to be used as character width.
                    add_width = vector_char_width
                    offsetX = minX-offset_x #overwrite offsetX with difference of starting X(minX) offset_x.
                    if use_custom: custom_width = add_width * custom_scale
										
                if kerning_vector >= 0 and use_custom: 
                    NumberOfStrokes_in_Path,Char_Vector_StrokeIDs_List = pdb.gimp_vectors_get_strokes(kerning_vector)
                    Char_Vector_Stroke = Char_Vector_StrokeIDs_List[0]
                    TypeOfStrokeInChar_Vector,NumberOfPointsInChar_Vector,ListOfControlPoints,IsStrokeClosed = pdb.gimp_vectors_stroke_get_points(kerning_vector,Char_Vector_Stroke)
                    firstX = ListOfControlPoints[2]
                    secondX = ListOfControlPoints[len(ListOfControlPoints)-4]
                    minX = min(firstX,secondX)
                    maxX = max(firstX,secondX)
                    kerning_width = maxX - minX #overwrite char_width to be used as character width.
                    add_width = kerning_width
                    offsetX = minX-offset_x #overwrite offsetX with difference of starting X(minX) offset_x.
                    if use_custom: custom_width = add_width * custom_scale
                    if use_custom: char_width = add_width * custom_scale # Slab v 0.06 caters for negative kerning with custom fonts
                    # Next line not necessary since the first character has to be displayed at the start of the user path
                    #if use_custom: f_xpos += (minX * custom_scale) # Slab v 0.06 caters for negative kerning with custom fonts
					

                dist =  f_xpos + float(char_width)/2 #calculate center dist of letter before adding to width.
                # Get the x,y position of the required point along the active path to position the characters
                x_point,y_point,SlopeAtThisPoint,IsThisPointValid = pdb.gimp_vectors_stroke_get_point_at_dist(active_vectors,ActivePath_StrokeID,dist,0.01)
                x_point1,y_point1,SlopeAtThisPoint1,IsThisPointValid = pdb.gimp_vectors_stroke_get_point_at_dist(active_vectors,ActivePath_StrokeID,dist-5,0.01)
                x_point2,y_point2,SlopeAtThisPoint2,IsThisPointValid = pdb.gimp_vectors_stroke_get_point_at_dist(active_vectors,ActivePath_StrokeID,dist+5,0.01)

                translate_x = x_point - float(char_width)/2
                translate_y = y_point - float(layer_id_list[layer_and_path_loop].height)/2 # ie on the path
                pdb.gimp_layer_translate(layer_id_list[layer_and_path_loop],translate_x - f_xpos,translate_y - text_layer_y_origin)
                
                if rotate_text:
                    if jitter: 
                        user_angle = fn_angle_jitter(original_user_angle,jitter_range)
                        angle = math.radians(user_angle)
                    else:
                        angle = math.radians(original_user_angle)
                else:
                    if jitter:
                        angle = computeThetaWithSlope([x_point1,y_point1], [x_point2,y_point2],SlopeAtThisPoint)
                        jitter_angle = math.degrees(angle) 
                        jitter_angle = fn_angle_jitter(jitter_angle,jitter_range)
                        angle = math.radians(jitter_angle)
                    else:
                        angle = computeThetaWithSlope([x_point1,y_point1], [x_point2,y_point2],SlopeAtThisPoint)

                # Initialise angle variables to be used in trying to compensate for distortions along curved path
                current_angle, prev_angle = math.degrees(angle), math.degrees(angle)
				
                x,y = pdb.gimp_drawable_offsets(layer_id_list[layer_and_path_loop])
                if reversed_writing:
                    if rotate_text:
                        rotated_angle = angle
                    else:
                        rotated_angle = angle + math.radians(180)
						
                    if Left_to_Right_path:
                        temp = pdb.gimp_item_transform_flip_simple(layer_id_list[layer_and_path_loop], 0, True, 0) # flips each character through 180 degrees on horzontal axis
                        temp = pdb.gimp_item_transform_rotate(temp,rotated_angle ,False,x+float(char_width)/2,y+float(temp.height/2))# then rotates each character by further 180 degrees
                    else:
                        temp = pdb.gimp_item_transform_flip_simple(layer_id_list[layer_and_path_loop], 0, True, 0) # flips each character through 180 degrees on horzontal axis
                        temp = pdb.gimp_item_transform_rotate(temp,rotated_angle ,False,x+float(char_width)/2,y+float(temp.height/2))# then rotates each character by further 180 degrees
                        
                else:
                    temp = pdb.gimp_item_transform_rotate(layer_id_list[loop],angle,False,x+float(char_width)/2,y+float(layer_id_list[layer_and_path_loop].height/2))
                
                curve_compensation = fn_compensate_for_curve_in_path(prev_angle, current_angle, f_size)
				
                if (keep_paths == 0):
                    if char_vector >=0 and not use_custom:
                        pdb.gimp_image_remove_vectors(image,path_id_list[layer_and_path_loop])
                layer_and_path_loop += 1

            else: # loop is greater than 0 i.e. there are subsequent characters to deal with
			
                if use_custom:
                    # get id of character from custom font file in memory
                    char_layer = pdb.gimp_image_get_layer_by_name(fontimg, custom_layer_name)
                    char_vector = pdb.gimp_image_get_vectors_by_name(fontimg,custom_layer_name)
                    if char_layer < 0: #if letter not found try swapping upper/lower case to look for
                        char = CurrentChar
                        if char.lower() == char:
                            char = char.upper()
                        else:
                            char = char.lower()
                        custom_layer_name = char + ".png"
                        char_layer = pdb.gimp_image_get_layer_by_name(fontimg, custom_layer_name)
                    # work out scale of character 
                    custom_scale = fn_get_custom_font_scale(f_size, char_layer.height)
                    # create, insert and set visibility the copy of the character layer into our image
                    copy = pdb.gimp_layer_new_from_drawable(char_layer,image)
                    pdb.gimp_image_insert_layer(image,copy,None,0)
                    pdb.gimp_item_set_visible(copy,True)
                    # sort out scaling to get the font size we need
                    custom_width = copy.width * custom_scale
                    custom_height = copy.height * custom_scale
                    old_interpolation = pdb.gimp_context_get_interpolation() # Note current interpolation method
                    pdb.gimp_context_set_interpolation(3) # Ensure that interpolation is set to Sinc (Lancos 3) method
                    pdb.gimp_layer_scale(copy,custom_width,custom_height,False)
                    pdb.gimp_context_set_interpolation(old_interpolation) # Restore old interpolation method
                    pdb.gimp_layer_set_offsets(copy,0,0)
                    # Add further layer identifiers to the list of text layer identifiers
                    layer_id_list.append(copy)
                    custom_scale = fn_get_custom_font_scale(f_size, char_layer.height)
                    char_width = custom_width
                    # Set the new custom layer colour from the current font colour and the colour to shade to
                    if spread_colour and not use_gradient_colour: 
                        if not use_space_to_change_step_size:
                            custom_colour = fn_shade_text_colour(original_font_colour, bl_colour, len(text)-1, loop)
                        else:
                            custom_colour = fn_shade_text_colour(original_font_colour, bl_colour, spaces_count - 1, shading_loop)
                        proc_colour_character(image, layer_id_list[layer_and_path_loop], custom_colour, custom_font_mode)
                    # Set the new custom layer colour from the current gradient sampled colour
                    if use_gradient_colour:
                        custom_colour = fn_shade_gradient_colour(colours_of_samples,loop)
                        proc_colour_character(image, layer_id_list[layer_and_path_loop], custom_colour, custom_font_mode)
					
                else:
                    # Add further layer identifiers to the list of text layer identifiers
                    layer_id_list.append(["text_layer" + str(layer_and_path_loop)])
                    # and create each subsequent layer
                    layer_id_list[layer_and_path_loop] = pdb.gimp_text_layer_new(image,CurrentChar,f_name,f_size,0)
				
                    # Insert subsequent layers into the image
                    pdb.gimp_image_insert_layer(image,layer_id_list[layer_and_path_loop],None,-1)
				
                    # Set the new text layer colour
                    if spread_colour and not use_gradient_colour: 
                        if not use_space_to_change_step_size:
                            f_colour = fn_shade_text_colour(original_font_colour, bl_colour, len(text)-1, loop)
                        else:
                            f_colour = fn_shade_text_colour(original_font_colour, bl_colour, spaces_count - 1, shading_loop)
                    # Set the new text layer colour from the current gradient sampled colour
                    if use_gradient_colour: 
                        if use_space_to_change_step_size: 
                            f_colour = fn_shade_gradient_colour(colours_of_samples,shading_loop)
                        else:
                            f_colour = fn_shade_gradient_colour(colours_of_samples,loop)
						
                    pdb.gimp_text_layer_set_color(layer_id_list[layer_and_path_loop], f_colour)
				
                    # Get width of current character
                    char_width = layer_id_list[layer_and_path_loop].width
                # Set layer opacity as per user setting
                pdb.gimp_layer_set_opacity(layer_id_list[layer_and_path_loop], tranparency)
                # Rename the layer
                pdb.gimp_item_set_name(layer_id_list[layer_and_path_loop],str(text[loop]))
                # Get the name of the current layer
                if use_custom: pdb.gimp_item_set_name(layer_id_list[layer_and_path_loop],str(text[loop]))
                temp = pdb.gimp_item_get_name((layer_id_list[layer_and_path_loop]))

                # Add further layer names to the list of text layer names
                layer_name_list.append(temp)

                # Rename the layer to account for multiple occurences of the same letter
                pdb.gimp_item_set_name(layer_id_list[layer_and_path_loop],str(layer_name_list[layer_and_path_loop]))
                # Set x,y the position of subsequent layers in the image
				
                pdb.gimp_layer_translate(layer_id_list[layer_and_path_loop],f_xpos,text_layer_y_origin )

                if not use_custom:
                    # Add further path identifiers to the list of path identifiers
                    path_id_list.append(pdb.gimp_vectors_new_from_text_layer(image,layer_id_list[layer_and_path_loop]))
                    # Add further names of the list paths names
                    # Use loop instead of layer_and_path_loop since we are referencing the original string for the path layer 
                    paths_names_list.append(str(joining_text_list[loop]))
                    # Insert path
                    temp = pdb.gimp_image_insert_vectors(image,path_id_list[layer_and_path_loop],None,-1)
				
                    char_vector = pdb.gimp_image_get_vectors_by_name(image, paths_names_list[layer_and_path_loop])
	
                kerning_vector = -1
                if use_custom:
                    char, n_char = CurrentChar, NextChar
                    char_layer = pdb.gimp_image_get_layer_by_name(fontimg, char + ".png")
                    if char_layer == None:
                        if char.lower() == char:
                            char = char.upper()
                        else:
                            char = char.lower()
							
                    if n_char<>"":
                        n_char_layer = pdb.gimp_image_get_layer_by_name(fontimg, n_char + ".png")
                        if n_char_layer < 0: #if letter not found try swapping upper/lower case to look for
                            if n_char.lower() == n_char:
                                n_char = n_char.upper()
                            else:
                                n_char = n_char.lower()

                    kerning_vector = pdb.gimp_image_get_vectors_by_name(fontimg,char + n_char + ".png") # check for character kerning pair paths with .png
                    if kerning_vector<0:
                        kerning_vector = pdb.gimp_image_get_vectors_by_name(fontimg,char + n_char) # check for character kerning pair paths without .png
					
					
                if char_vector >=0 and kerning_vector <0:

                    char_vector_name = pdb.gimp_item_get_name(char_vector)
                    NumberOfStrokes_in_Path,Char_Vector_StrokeIDs_List = pdb.gimp_vectors_get_strokes(char_vector)
                    Char_Vector_Stroke = Char_Vector_StrokeIDs_List[0]
                    TypeOfStrokeInChar_Vector,NumberOfPointsInChar_Vector,ListOfControlPoints,IsStrokeClosed = pdb.gimp_vectors_stroke_get_points(char_vector,Char_Vector_Stroke)
                    firstX = ListOfControlPoints[2]

                    secondX = ListOfControlPoints[len(ListOfControlPoints)-4]
                    minX = min(firstX,secondX)
                    maxX = max(firstX,secondX)
                    vector_char_width = maxX - minX #overwrite char_width to be used as character width.
                    add_width = vector_char_width
                    offsetX = minX-offset_x #overwrite offsetX with difference of starting X(minX) offset_x.
                    if use_custom: custom_width = add_width * custom_scale
                    if use_custom: char_width = add_width * custom_scale # Slab v 0.06 caters for negative kerning with custom fonts
                    if use_custom: f_xpos += (minX * custom_scale) # Slab v 0.06 caters for negative kerning with custom fonts


                if kerning_vector >= 0: 
                    NumberOfStrokes_in_Path,Char_Vector_StrokeIDs_List = pdb.gimp_vectors_get_strokes(kerning_vector)
                    Char_Vector_Stroke = Char_Vector_StrokeIDs_List[0]
                    TypeOfStrokeInChar_Vector,NumberOfPointsInChar_Vector,ListOfControlPoints,IsStrokeClosed = pdb.gimp_vectors_stroke_get_points(kerning_vector,Char_Vector_Stroke)
                    firstX = ListOfControlPoints[2]
                    secondX = ListOfControlPoints[len(ListOfControlPoints)-4]
                    minX = min(firstX,secondX)
                    maxX = max(firstX,secondX)
                    kerning_width = maxX - minX #overwrite char_width to be used as character width.
                    add_width = kerning_width
                    offsetX = minX-offset_x #overwrite offsetX with difference of starting X(minX) offset_x.
                    if use_custom: custom_width = add_width * custom_scale
                    if use_custom: char_width = add_width * custom_scale # Slab v 0.06 caters for negative kerning with custom fonts
                    if use_custom: f_xpos += (minX * custom_scale) # Slab v 0.06 caters for negative kerning with custom fonts
					
                					
                dist =  f_xpos + float(char_width)/2 #calculate center dist of letter before adding to width.

                # Get the x,y position of the required point along the active path to position the characters
                x_point,y_point,SlopeAtThisPoint,IsThisPointValid = pdb.gimp_vectors_stroke_get_point_at_dist(active_vectors,ActivePath_StrokeID,dist,0.01)
                #x_point1,y_point1,SlopeAtThisPoint1,IsThisPointValid = pdb.gimp_vectors_stroke_get_point_at_dist(active_vectors,ActivePath_StrokeID,dist-5,0.01)
                #x_point2,y_point2,SlopeAtThisPoint2,IsThisPointValid = pdb.gimp_vectors_stroke_get_point_at_dist(active_vectors,ActivePath_StrokeID,dist+5,0.01)
                x_point1,y_point1,SlopeAtThisPoint1,IsThisPointValid = pdb.gimp_vectors_stroke_get_point_at_dist(active_vectors,ActivePath_StrokeID,dist,0.01)
                x_point2,y_point2,SlopeAtThisPoint2,IsThisPointValid = pdb.gimp_vectors_stroke_get_point_at_dist(active_vectors,ActivePath_StrokeID,dist+0.5,0.01)
                if rotate_text:
                    if jitter: 
                        user_angle = fn_angle_jitter(original_user_angle,jitter_range)
                        angle = math.radians(user_angle)
                    else:
                        angle = math.radians(original_user_angle)
                else:
                    if jitter:
                        angle = computeThetaWithSlope([x_point1,y_point1], [x_point2,y_point2],SlopeAtThisPoint)
                        jitter_angle = math.degrees(angle) 
                        jitter_angle = fn_angle_jitter(jitter_angle,jitter_range)
                        angle = math.radians(jitter_angle)
                    else:
                        angle = computeThetaWithSlope([x_point1,y_point1], [x_point2,y_point2],SlopeAtThisPoint)

                #angle = computeThetaWithSlope([x_point1,y_point1], [x_point2,y_point2],SlopeAtThisPoint)
                #angle = computeThetaWithSlope([x_point,y_point], [x_point,y_point],SlopeAtThisPoint)
                current_angle = math.degrees(angle)
								
                curve_compensation,temp_angle_diff = fn_compensate_for_curve_in_path(prev_angle, current_angle, f_size)
                curve_compensation = (curve_compensation * compensate_multiplier)
                curve = curve_compensation
				
                dist =  (f_xpos + curve_compensation) + float(char_width)/2 #calculate center dist of letter before adding to width.

                x_point,y_point,SlopeAtThisPoint,IsThisPointValid = pdb.gimp_vectors_stroke_get_point_at_dist(active_vectors,ActivePath_StrokeID,dist,0.01)
                x_point1,y_point1,SlopeAtThisPoint1,IsThisPointValid = pdb.gimp_vectors_stroke_get_point_at_dist(active_vectors,ActivePath_StrokeID,dist-5,0.01)
                x_point2,y_point2,SlopeAtThisPoint2,IsThisPointValid = pdb.gimp_vectors_stroke_get_point_at_dist(active_vectors,ActivePath_StrokeID,dist+5,0.01)
				
                translate_x = x_point - float(char_width)/2
                if char_vector>=0 and use_custom: translate_x += (minX * custom_scale)  # Slab v 0.06 cater for negative kerning with custom fonts
                translate_y = y_point - float(layer_id_list[layer_and_path_loop].height)/2 # ie on the path
                pdb.gimp_layer_translate(layer_id_list[layer_and_path_loop],translate_x - f_xpos,translate_y - text_layer_y_origin)
												
                x,y = pdb.gimp_drawable_offsets(layer_id_list[layer_and_path_loop])
				
                if reversed_writing:
                    if rotate_text:
                        rotated_angle = angle
                    else:
                        rotated_angle = angle + math.radians(180)
						
                    if Left_to_Right_path:
						
                        temp = pdb.gimp_item_transform_flip_simple(layer_id_list[layer_and_path_loop], 0, True, 0) # flip original layer along horizontal axis
                        temp = pdb.gimp_item_transform_rotate(temp,rotated_angle ,False,x+float(char_width)/2,y+float(temp.height/2)) # then rotate flipped layer on path with extra 180 degrees
                    else:
                        temp = pdb.gimp_item_transform_flip_simple(layer_id_list[layer_and_path_loop], 0, True, 0) # flip original layer along horizontal axis
                        temp = pdb.gimp_item_transform_rotate(temp,rotated_angle ,False,x+float(char_width)/2,y+float(temp.height/2)) # then rotate flipped layer on path with extra 180 degrees
                else:
                    #v 2.11 line
                    temp = pdb.gimp_item_transform_rotate(layer_id_list[layer_and_path_loop],angle,False,x+float(char_width)/2,y+float(layer_id_list[layer_and_path_loop].height/2)) # rotate original layer on path
						
                if not use_custom:						
                    # Delete path
                    if (keep_paths == 0):
                        pdb.gimp_image_remove_vectors(image,path_id_list[layer_and_path_loop])
                # Merge new character layer to first character layer
                if (keep_layers == 0):
                    pdb.gimp_image_merge_down(image, layer_id_list[layer_and_path_loop],0)
                temp = pdb.gimp_displays_flush()

                layer_and_path_loop += 1				

        # Update y position adjustment
        if use_custom:
            text_layer_y_origin = text_layer_y_origin + (font_y_offset - (custom_height / 2))
    	    font_y_offset = (custom_height / 2)
            f_xpos = f_xpos + custom_width + kern_gap + curve - xpos_adjustment
        else:
            (fontbox_width, fontbox_height, fontbox_ascent, fontbox_descent) = pdb.gimp_text_get_extents_fontname(CurrentChar, f_size, PIXELS, str(f_name))
            text_layer_y_origin = text_layer_y_origin + (font_y_offset - fontbox_ascent)
            font_y_offset = fontbox_ascent
		
            f_xpos = f_xpos + fontbox_width + kern_gap + curve - xpos_adjustment
		
        prev_angle = current_angle

        if (CurrentChar == " "):
            
            f_size_adjustment = step_size
            f_size = f_size + f_size_adjustment
            # Stop font size becoming too small/large
            if (f_size<2):
                f_size = 2
                governor += 1
                if governor > 2: loop = len(joining_text_list)
            if (f_size>3000):
                f_size = 3000
                governor += 1
                if governor > 2: loop = len(joining_text_list)
				
            if not use_custom: kern_gap += (f_size * kern_factor) # otherwise it will double the amount of padding for the custom font
            shading_loop += 1
			
            if use_text_size_modifier: # check to see if font size modifier is being used to begin changing size
                if use_space_to_change_step_size:
				
                    if attack == spaces_loop:
                        decay = target - attack
                        if not even: decay -= 1
						
                        if decay < 1: decay = 1
                        size_difference,new_step_size = 0,0
						
                        if medial_text_size > final_text_size: size_difference = medial_text_size - final_text_size
                        if final_text_size > medial_text_size: size_difference = final_text_size - medial_text_size
						
                        if target>3:
                            new_step_size = size_difference/decay
                        else:
                            if decay>1:
                                new_step_size = size_difference/(decay-1)
                            else:
                                new_step_size = size_difference/(decay)
							
                        if even: 
                            if final_text_size > medial_text_size:
						        f_size = medial_text_size + new_step_size                                
                            else:
                                f_size -= (step_size + (new_step_size * 1))
						   
                        if not even: f_size -= step_size
						
                        if final_text_size < medial_text_size: new_step_size = 0 - new_step_size
                        step_size = new_step_size
						
                    spaces_loop += 1
					
                else:
				
                    if attack == loop:							
                        decay = target - attack
                        if even: 
                            decay -=2 
                        else:
                            decay -= 0							
                        if decay < 1: decay = 1
                        size_difference,new_step_size = 0,0
						
                        if medial_text_size > final_text_size: size_difference = medial_text_size - final_text_size
                        if final_text_size > medial_text_size: size_difference = final_text_size - medial_text_size
						
                        if target>3:
                            if decay>1:
                                if even:
                                    new_step_size = size_difference/decay
                                else:
                                    new_step_size = size_difference/(decay-1)
                            else:
                                new_step_size = size_difference/decay
                        else:
                            if decay>1:
                                new_step_size = size_difference/(decay-1)
                            else:
                                new_step_size = size_difference/decay
							
                        if not even: 
                            if final_text_size > medial_text_size:
						        f_size = medial_text_size + new_step_size
                                
                            else:
                                f_size -= (step_size + (new_step_size * 1))
						   
                        if even: f_size -= (step_size)
							
                        if final_text_size < medial_text_size: new_step_size = 0 - new_step_size
                        step_size = new_step_size
        else:
            if (use_space_to_change_step_size == 0):
                f_size_adjustment = step_size
                f_size = f_size + f_size_adjustment
                if (f_size<2):
                    f_size = 2
                    governor += 1
                    if governor > 2: loop = len(joining_text_list)
                if (f_size>3000):
                    f_size = 3000
                    governor += 1
                    if governor > 2: loop = len(joining_text_list)
                if use_text_size_modifier: # check to see if font size modifier is being used to begin changing size
                    if attack == loop:
                        decay = len(text) - attack
                        if even: 
                            decay -=2 
                        else:
                            decay -= 0
							
                        if decay < 1: decay = 1
                        size_difference,new_step_size = 0,0
						
                        if medial_text_size > final_text_size: size_difference = medial_text_size - final_text_size
                        if final_text_size > medial_text_size: size_difference = final_text_size - medial_text_size
						
                        if len(text)>3:
                            if decay>1:
                                if even:
                                    new_step_size = size_difference/decay
                                else:
                                    new_step_size = size_difference/(decay-1)
                            else:
                                new_step_size = size_difference/decay
                        else:
                            if decay>1:
                                new_step_size = size_difference/(decay-1)
                            else:
                                new_step_size = size_difference/decay
							
                        if not even: 
                            if final_text_size > medial_text_size:
						        f_size = medial_text_size + new_step_size
                                
                            else:
                                f_size -= (step_size + (new_step_size * 1))
						   
                        if even: f_size -= (step_size)
						
                        if (f_size<2): f_size = 2
                        if (f_size>3000): f_size = 3000
                        if final_text_size < medial_text_size: new_step_size = 0 - new_step_size
                        step_size = new_step_size
                kern_gap = (f_size * kern_factor)
        
		
		#### New text added to take account of adjusted font size for next character
        if use_text_size_modifier:
            if not even: 
                if final_text_size > medial_text_size:
                    temp_f_size = f_size + new_step_size
                else:
                    temp_f_size = f_size - new_step_size
						   
            else:
                if final_text_size > medial_text_size:
                    temp_f_size = f_size + new_step_size
                else:
                    temp_f_size = f_size - new_step_size
				
            if (temp_f_size<2): temp_f_size = 2
            if (temp_f_size>3000): temp_f_size = 3000
            (tempbox_width, tempbox_height, tempbox_ascent, tempbox_descent) = pdb.gimp_text_get_extents_fontname(NextChar, temp_f_size, PIXELS, str(f_name))
                
        else:
            if use_custom: 
                tempbox_width = 0
                if not do_best_size:
                    if NextChar <> "":
                        tempbox_layer_name = CurrentChar + ".png"
                        if CurrentChar == " ":
                            tempbox_width = space_width
                        else:
                            tempbox_layer = pdb.gimp_image_get_layer_by_name(fontimg,tempbox_layer_name)
                            if tempbox_layer<0: # Added 30/03
                                char = CurrentChar # Added 30/03
                                if char.lower() == char: # Added 30/03
                                    char = char.upper() # Added 30/03
                                else: # Added 30/03
                                    char = char.lower() # Added 30/03
                                tempbox_layer_name = char + ".png" # Added 30/03
                                tempbox_layer = pdb.gimp_image_get_layer_by_name(fontimg,tempbox_layer_name) # Added 30/03
							
                            custom_scale = fn_get_custom_font_scale(f_size, tempbox_layer.height)
                            tempbox_width = tempbox_layer.width * custom_scale
                    
            else:
                (tempbox_width, tempbox_height, tempbox_ascent, tempbox_descent) = pdb.gimp_text_get_extents_fontname(NextChar, f_size, PIXELS, str(f_name))
								
        NextChr_width = tempbox_width
        loop = loop + 1
		
    if (keep_layers == 0):
        final_layer = pdb.gimp_image_get_active_layer(image)
        label_repeat_info = ""
        if repeat_text: label_repeat_info = " repeat"
        label_colour_info = ""
        if spread_colour: label_colour_info = " shaded"
        if use_gradient_colour: label_colour_info = " gradient used"
        if initial_step_size <> 0: label_fontsize_info += " with " + str(initial_step_size) + "pt steps"
        if use_text_size_modifier:
            label_fontsize_info = " " + str(int(initial_text_size)) + "pt" + " to " + str(int(medial_text_size)) + "pt" + " to "  + str(int(final_text_size)) + "pt"
        label_details += label_repeat_info
        label_details += label_colour_info
        label_details += label_fontsize_info
        if use_space_to_change_step_size:
            size_method = " by word"
        else:
            size_method = " by char"
        label_details += size_method
        label_details += " Font: " + label_fontname
        if reversed_writing: 
            label_reverse =  " reversed"
            label_details += label_reverse
        if rotate_text:
            label_rotate = " rotated " + str(user_angle) + str(unichr(176))
            label_details += label_rotate
        if (jitter and jitter_range <> 0):
            label_jitter = " jitter +/- " + str(jitter_range/2) + str(unichr(176))
            label_details += label_jitter
        
        pdb.gimp_item_set_name(final_layer,label_details)
		
        
        if reversed_vector > -1: pdb.gimp_image_remove_vectors(image, reversed_vector)
	
##########################################################################################################

#### End of proc_text_along_path
 	
 	
#####################################################################################################

def proc_dropshadow(im, target_layer, drop_offset_x, drop_offset_y, drop_blur, drop_opacity, drop_colour):
    #REM Create a text drop shadow effect - optionally clipped to the text layer for transparent type texts
	
    # make a copy of the image width/height
    old_im_width, old_im_height = im.width, im.height
    name = pdb.gimp_item_get_name(target_layer)
    # Copy layer to use as the drop shadow
    shadow_layer = pdb.gimp_layer_copy(target_layer, False)
    # Add and re-position the layer in the layer stack
    im.add_layer(shadow_layer, -1)
    pdb.gimp_image_lower_item(im, shadow_layer)
    # get layer offsets for the shadow layer (which will also be the same as the original layer holding the user text)
    shadow_offset_x, shadow_offset_y = pdb.gimp_drawable_offsets(shadow_layer)
    # work out the centre point of the shadow/original layer
    shad_centre_x = (shadow_layer.width/2)
    shad_centre_y = (shadow_layer.height/2)
    # make up new size for the shadow layer to accommodate blurred result - currently twice the existing width and height
    new_shad_width = shadow_layer.width*2
    new_shad_height = shadow_layer.height*2
    # work out new offsets for the new size of the shadow layer to ensure that the shadow and original layers are still aligned
    new_shad_layer_offset_x = 0 - ((shadow_layer.width - new_shad_width)/2)
    new_shad_layer_offset_y = 0 - ((shadow_layer.height - new_shad_height)/2)
    # and apply the new offsets to the shadow layer
    pdb.gimp_layer_resize(shadow_layer, new_shad_width, new_shad_height, new_shad_layer_offset_x, new_shad_layer_offset_y)
    # make a note of the new shadow layer offsets
    shadow_resized_offset_x, shadow_resized_offset_y = pdb.gimp_drawable_offsets(shadow_layer)	
	
    # if any of the shadow layer falls outside the image boundaries
    reset_shadow = False
    if new_shad_layer_offset_x<0 or new_shad_layer_offset_y<0 or shadow_layer.width > im.width or shadow_layer.height > im.height:
        reset_shadow = True
        # set offsets of the shadow layer to image origin
        pdb.gimp_layer_set_offsets(shadow_layer,0,0)
        # set offsets for original text layer to match
        pdb.gimp_layer_set_offsets(target_layer,0 + abs(new_shad_layer_offset_x) ,0 + abs(new_shad_layer_offset_y))
        # resize the image to accomodate the new shadow layer size
        new_im_width, new_im_height = im.width, im.height
        if shadow_layer.width > im.width: new_im_width = shadow_layer.width
        if shadow_layer.height > im.height: new_im_height = shadow_layer.height
        pdb.gimp_image_resize(im, new_im_width, new_im_height, 0, 0)
		
    # get alpha selection
    pdb.gimp_image_select_item(im, 2, shadow_layer)
    # fill the text on the shadow layer with the background colour
    if drop_colour<3:
        pdb.gimp_drawable_edit_fill(shadow_layer,drop_colour-1)
    # deselect
    pdb.gimp_selection_none(im)
    # blur shadow 
    pdb.plug_in_gauss(im, shadow_layer, drop_blur, drop_blur, 1)
    # crop the blurred layer
    pdb.plug_in_autocrop_layer(im, shadow_layer)
    # move the shadow layer to user position and adjust opacity
    temp = pdb.gimp_layer_translate(shadow_layer, drop_offset_x, drop_offset_y)
    pdb.gimp_layer_set_opacity(shadow_layer, drop_opacity)
    # merge and rename
    target_layer = pdb.gimp_image_merge_down(im,target_layer,0)
    pdb.gimp_item_set_name(target_layer, name)
    if reset_shadow: 
        cropped_offset_x, cropped_offset_y = pdb.gimp_drawable_offsets(target_layer)	
        pdb.gimp_layer_set_offsets(target_layer,shadow_offset_x, shadow_offset_y)
    pdb.gimp_image_resize(im, old_im_width, old_im_height, 0, 0)
    

    #### Other Useful Functions
# -------------------------------------------------------------------------------------
	
def fn_selection_info(im):
    # returns selection info
    selection,x1,y1,x2,y2 = pdb.gimp_selection_bounds(im)
    s_width, s_height = x2 - x1, y2 - y1
	
    return selection, x1, x2, y1, y2, s_width, s_height


######################################################################################################################################
	
def fn_selection_amount(sel_value,r_ness,adjustment):
    # 
    value = (math.sqrt(sel_value)*r_ness)*adjustment
    return value

######################################################################################################################################
	
def proc_messagebox(message):
    old_handler = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2)
    pdb.gimp_message(str(message))
    pdb.gimp_message_set_handler(old_handler)
	
######################################################################################################################################
		
def proc_trim_away(img, layer, amount):
    #### Trim off pixels from a layer
    # set our feather settings
    our_feather = True
    our_feather_radius = 2
    #get feather settings
    feather = pdb.gimp_context_get_feather()
    feather_x, feather_y = pdb.gimp_context_get_feather_radius()
    # make sure that feather is set
    if not feather: pdb.gimp_context_set_feather(our_feather)
    # use our feather radius
    pdb.gimp_context_set_feather_radius(our_feather_radius, our_feather_radius)
    # select the contents of the layer
    pdb.gimp_image_select_item(img, 2, layer)
    # shrink the selection
    pdb.gimp_selection_shrink(img, amount)
    # invert selection
    pdb.gimp_selection_invert(img)
    # trim away pixels
    successful = pdb.gimp_edit_cut(layer)
    # deselect
    pdb.gimp_selection_none(img)
    # reset feather settings
    pdb.gimp_context_set_feather(feather)
    pdb.gimp_context_set_feather_radius(feather_x, feather_y)
	
    #### End of Proc_trim_away

######################################################################################################################################
	
register(
    "python_fu_Custom_Font_Paragrapher_v009",
    "Pat's simple paragraph writer for use with Custom Fonts",
    "Pat's simple paragraphs writer for use with Custom Fonts",
    "Skinnyhouse",
    "Open source (BSD 3-clause license)",
    "2020",
    "<Image>/Filters/Custom Font Tools/Custom Font Paragrapher v0.009...",
    "RGB*",
    [
    (PF_FILENAME, "pf_names_xcffilename", ".xcf custom font file: ", 0),
    (PF_TEXT, "pf_string", "Text: ", "Gimp is fun."),
    (PF_SLIDER, "pf_fontsize", "Font size: ", 100, (8, 600, 1)),
    (PF_SLIDER, "pf_linespace", "Line spacing amount: ", 0, (-75, 500, 1)),
    (PF_SLIDER, "pf_top", "Top margin amount: ", 50, (0, 250, 1)),
    (PF_SLIDER, "pf_left", "Left margin amount: ", 50, (0, 250, 1)),
    (PF_SLIDER, "pf_right", "Right margin amount: ", 50, (0, 250, 1)),
    (PF_SLIDER, "pf_bottom", "Bottom margin amount: ", 50, (0, 250, 1)),
    (PF_OPTION, "pf_indent_option", "Indent options", 0, ("No indents", "Indent new paragraphs")),
    (PF_SLIDER, "pf_indent_size", "Number of spaces widths for indent: ", 2, (1, 8, 1)),
    (PF_OPTION, "bg_option", "Background options: ", 0, ("None", "Use foreground colour", "Use background colour")),
    (PF_SLIDER, "pf_drop_offset_x", "Shadow offset x position: ", 8, (-50, 50, 1)),
    (PF_SLIDER, "pf_drop_offset_y", "Shadow offset y position: ", 8, (-50, 50, 1)),
    (PF_SLIDER, "pf_drop_blur", "Shadow blur percentage: ", 15, (0, 50, 1)),
    (PF_SLIDER, "pf_drop_opacity", "Shadow opacity: ", 70, (20, 100, 1)),
    (PF_OPTION, "drop_colour", "Drop shadow", 0, ( "No Shadow", "Foreground Colour", "Background colour", "Layer copy")),
    (PF_TOGGLE, "copy_paths", "Use active path as line template\n(if available): ", 0),
    (PF_TOGGLE, "centre_layers", "Justify layers to image centre: ", 0),
    ],
    [],
    do_stuff)

main()
